# -*- coding: utf-8 -*-
"""
Resource extractors
"""
from .resource_extraction import (ResourceX, MultiFileResourceX,
                                  MultiTimeResourceX, MultiYearResourceX,
                                  NSRDBX, MultiFileNSRDBX,
                                  MultiTimeNSRDBX, MultiYearNSRDBX,
                                  WindX, MultiFileWindX,
                                  MultiTimeWindX, MultiYearWindX,
                                  WaveX, MultiTimeWaveX, MultiYearWaveX)
from .temporal_stats import TemporalStats



# -*- coding: utf-8 -*-
"""
Execution utilities.
"""
from abc import ABC, abstractmethod
import subprocess
import logging
import getpass
import shlex
from warnings import warn

from rex.utilities.execution import SubprocessManager
from rex.utilities.exceptions import (ExecutionError, HpcError, SlurmWarning,
                                      PbsWarning)


logger = logging.getLogger(__name__)


class HpcJobManager(SubprocessManager, ABC):
    """Abstract HPC job manager framework"""

    # get username as class attribute.
    USER = getpass.getuser()

    # HPC queue column headers
    QCOL_NAME = None  # Job name column
    QCOL_ID = None  # Job integer ID column
    QCOL_STATUS = None  # Job status column

    # set a max job name length, will raise error if too long.
    MAX_NAME_LEN = 100

    # default rows to skip in queue stdout
    QSKIP = None

    def __init__(self, user=None, queue_dict=None):
        """
        Parameters
        ----------
        user : str | None
            HPC username. None will get your username using getpass.getuser()
        queue_dict : dict | None
            Parsed HPC queue dictionary (qstat for PBS or squeue for SLURM)
            from parse_queue_str(). None will get the queue from PBS or SLURM.
        """

        self._user = user
        if self._user is None:
            self._user = self.USER

        if queue_dict is not None and not isinstance(queue_dict, dict):
            emsg = ('HPC queue_dict arg must be None or Dict but received: '
                    '{}, {}'.format(queue_dict, type(queue_dict)))
            logger.error(emsg)
            raise HpcError(emsg)

        self._queue = queue_dict

    @staticmethod
    def _skip_q_rows(queue_str, skip_rows):
        """Remove rows from the queue_str that are to be skipped.

        Parameters
        ----------
        queue_str : str
            HPC queue output string. Can be split on line breaks to get list.
        skip_rows : int | list | None
            Optional row index values to skip.

        Returns
        -------
        queue_str : str
            HPC queue output string. Can be split on line breaks to get list.
        """
        if skip_rows is not None:
            if isinstance(skip_rows, int):
                skip_rows = [skip_rows]

            queue_str = [row for i, row in enumerate(queue_str.split('\n'))
                         if i not in skip_rows]
            queue_str = '\n'.join(queue_str)

        return queue_str

    @classmethod
    def parse_queue_str(cls, queue_str, keys=0):
        """Parse the qstat or squeue output string into a dict format keyed by
        integer job id with nested dictionary of job properties (queue
        printout columns).

        Parameters
        ----------
        queue_str : str
            HPC queue output string (qstat for PBS or squeue for SLURM).
            Typically a space-delimited string with line breaks.
        keys : list | int
            Argument to set the queue job attributes (column headers).
            This defaults to an integer which says which row index contains
            the space-delimited column headers. Can also be a list to
            explicitly set the column headers.

        Returns
        -------
        queue_dict : dict
            HPC queue parsed into dictionary format keyed by integer job id
            with nested dictionary of job properties (queue printout columns).
        """

        queue_dict = {}
        queue_rows = queue_str.split('\n')

        if isinstance(keys, int):
            del_index = keys
            keys = [k.strip(' ')
                    for k in queue_rows[keys].strip(' ').split(' ')
                    if k != '']
            del queue_rows[del_index]

        for row in queue_rows:
            job = [k.strip(' ') for k in row.strip(' ').split(' ') if k != '']
            job_id = int(job[keys.index(cls.QCOL_ID)])
            queue_dict[job_id] = {k: job[i] for i, k in enumerate(keys)}

        return queue_dict

    @abstractmethod
    def query_queue(self, job_name=None, user=None, qformat=None,
                    skip_rows=None):
        """Run the HPC queue command and return the raw stdout string.

        Parameters
        ----------
        job_name : str | None
            Optional to check the squeue for a specific job name (not limited
            to the 8 shown characters) or None to show user's whole queue.
        user : str | None
            HPC username. None will get your username using getpass.getuser()
        qformat : str | None
            Queue format string specification. Changing this form the
            default (None) could have adverse effects!
        skip_rows : int | list | None
            Optional row index values to skip.

        Returns
        -------
        stdout : str
            HPC queue output string. Can be split on line breaks to get list.
        """

    @property
    def queue(self):
        """Get the HPC queue parsed into dict format keyed by integer job id

        Returns
        -------
        queue : dict
            HPC queue parsed into dictionary format keyed by integer job id
            with nested dictionary of job properties (queue printout columns).
        """
        if self._queue is None:
            qstr = self.query_queue(user=self._user)
            self._queue = self.parse_queue_str(qstr)

        return self._queue

    @property
    def queue_job_names(self):
        """Get a list of the job names in the queue"""
        return [attrs[self.QCOL_NAME] for attrs in self.queue.values()]

    @property
    def queue_job_ids(self):
        """Get a list of the job integer ids in the queue"""
        return list(self.queue.keys())

    def check_status(self, job_id=None, job_name=None):
        """Check the status of an HPC job using the HPC queue.

        Parameters
        ----------
        job_id : int | None
            Job integer ID number (preferred input)
        job_name : str
            Job name string.

        Returns
        -------
        status : str | NoneType
            Queue job status str or None if not found.
            SLURM status strings: PD, R, CG (pending, running, complete).
            PBS status strings: Q, R, C (queued, running, complete).
        """

        status = None

        if job_id is not None:
            if int(job_id) in self.queue:
                status = self.queue[int(job_id)][self.QCOL_STATUS]

        elif job_name is not None:
            if job_name in self.queue_job_names:
                for attrs in self.queue.values():
                    if attrs[self.QCOL_NAME] == job_name:
                        status = attrs[self.QCOL_STATUS]
                        break

        else:
            msg = 'Need a job_id or job_name to check HPC job status!'
            logger.error(msg)
            raise HpcError(msg)

        return status


class PBS(HpcJobManager):
    """Subclass for PBS subprocess jobs."""

    # PBS qstat column headers
    QCOL_NAME = 'Name'  # Job name column
    QCOL_ID = 'Job id'  # Job integer ID column
    QCOL_STATUS = 'S'  # Job status column

    # Frozen PBS qstat column headers cached b/c its not space delimited
    QSTAT_KEYS = ('Job id', 'Name', 'User', 'Time Use', 'S', 'Queue')

    # set a max job name length, will raise error if too long.
    MAX_NAME_LEN = 100

    # default rows to skip in queue stdout
    QSKIP = (0, 1)

    def __init__(self, user=None, queue_dict=None):
        """
        Parameters
        ----------
        user : str | None
            HPC username. None will get your username using getpass.getuser()
        queue_dict : dict | None
            Parsed HPC queue dictionary (qstat for PBS or squeue for SLURM)
            from parse_queue_str(). None will get the queue from PBS or SLURM.
        """
        super().__init__(user=user, queue_dict=queue_dict)

    @classmethod
    def query_queue(cls, job_name=None, user=None, qformat=None,
                    skip_rows=None):
        """Run the PBS qstat command and return the raw stdout string.

        Parameters
        ----------
        job_name : str | None
            Optional to check the squeue for a specific job name (not limited
            to the 8 shown characters) or None to show user's whole queue.
        user : str | None
            HPC username. None will get your username using getpass.getuser()
        qformat : str | None
            Queue format string specification. Changing this form the
            default (None) could have adverse effects!
        skip_rows : int | list | None
            Optional row index values to skip.

        Returns
        -------
        stdout : str
            qstat output string. Can be split on line breaks to get list.
        """

        if user is None:
            user = cls.USER

        if skip_rows is None:
            skip_rows = cls.QSKIP

        cmd = 'qstat -u {user}'.format(user=user)
        stdout, _ = cls.submit(cmd)
        stdout = cls._skip_q_rows(stdout, skip_rows)
        return stdout

    @property
    def queue(self):
        """Get the HPC queue parsed into dict format keyed by integer job id

        Returns
        -------
        queue : dict
            HPC queue parsed into dictionary format keyed by integer job id
            with nested dictionary of job properties (queue printout columns).
        """
        if self._queue is None:
            qstr = self.query_queue(user=self._user)
            self._queue = self.parse_queue_str(qstr, keys=self.QSTAT_KEYS)

        return self._queue

    def qsub(self, cmd, alloc, queue, name='reV', feature=None,
             stdout_path='./stdout', keep_sh=False):
        """Submit a PBS job via qsub command and PBS shell script

        Parameters
        ----------
        cmd : str
            Command to be submitted in PBS shell script. Example:
                'python -m reV.generation.cli_gen'
        alloc : str
            HPC allocation account. Example: 'rev'.
        queue : str
            HPC queue to submit job to. Example: 'short', 'batch-h', etc...
        name : str
            PBS job name.
        feature : str | None
            PBS feature request (-l {feature}).
            Example: 'feature=24core', 'qos=high', etc...
        stdout_path : str
            Path to print .stdout and .stderr files.
        keep_sh : bool
            Boolean to keep the .sh files. Default is to remove these files
            after job submission.

        Returns
        -------
        out : str
            qsub standard output, this is typically the PBS job ID.
        err : str
            qsub standard error, this is typically an empty string if the job
            was submitted successfully.
        """

        if len(name) > self.MAX_NAME_LEN:
            msg = ('Cannot submit job with name longer than {} chars: "{}"'
                   .format(self.MAX_NAME_LEN, name))
            logger.error(msg)
            raise ValueError(msg)

        status = self.check_status(job_name=name)

        if status in ('Q', 'R'):
            logger.info('Not submitting job "{}" because it is already in '
                        'qstat with status: "{}"'.format(name, status))
            out = None
            err = 'already_running'
        else:
            self.make_path(stdout_path)
            feature_str = '#PBS -l {}\n'.format(str(feature).replace(' ', ''))
            fname = '{}.sh'.format(name)
            script = ('#!/bin/bash\n'
                      '#PBS -N {n} # job name\n'
                      '#PBS -A {a} # allocation account\n'
                      '#PBS -q {q} # queue (debug, short, batch, or long)\n'
                      '#PBS -o {p}/{n}_$PBS_JOBID.o\n'
                      '#PBS -e {p}/{n}_$PBS_JOBID.e\n'
                      '{L}'
                      'echo Running on: $HOSTNAME, Machine Type: $MACHTYPE\n'
                      '{cmd}'
                      .format(n=name, a=alloc, q=queue, p=stdout_path,
                              L=feature_str if feature else '',
                              cmd=cmd))

            # write the shell script file and submit as qsub job
            self.make_sh(fname, script)
            out, err = self.submit('qsub {script}'.format(script=fname))

            if not keep_sh:
                self.rm(fname)

            if err:
                msg = 'Received a PBS error or warning: {}'.format(err)
                logger.warning(msg)
                warn(msg, PbsWarning)
            else:
                logger.debug('PBS job "{}" with id #{} submitted successfully'
                             .format(name, out))
                self._queue[int(out)] = {self.QCOL_ID: int(out),
                                         self.QCOL_NAME: name,
                                         self.QCOL_STATUS: 'Q'}

        return out, err


class SLURM(HpcJobManager):
    """Subclass for SLURM subprocess jobs."""

    # SLURM squeue column headers
    QCOL_NAME = 'NAME'  # Job name column
    QCOL_ID = 'JOBID'  # Job integer ID column
    QCOL_STATUS = 'ST'  # Job status column

    MAX_NAME_LEN = 100
    SQ_FORMAT = ("%.15i %.30P  %.{}j  %.20u %.10t %.15M %.25R %q"
                 .format(MAX_NAME_LEN))

    # default rows to skip in queue stdout
    QSKIP = None

    def __init__(self, user=None, queue_dict=None):
        """
        Parameters
        ----------
        user : str | None
            HPC username. None will get your username using getpass.getuser()
        queue_dict : dict | None
            Parsed HPC queue dictionary (qstat for PBS or squeue for SLURM)
            from parse_queue_str(). None will get the queue from PBS or SLURM.
        """
        super().__init__(user=user, queue_dict=queue_dict)

    @classmethod
    def query_queue(cls, job_name=None, user=None, qformat=None,
                    skip_rows=None):
        """Run the HPC queue command and return the raw stdout string.

        Parameters
        ----------
        job_name : str | None
            Optional to check the squeue for a specific job name (not limited
            to the 8 shown characters) or None to show user's whole queue.
        user : str | None
            HPC username. None will get your username using getpass.getuser()
        qformat : str | None
            Queue format string specification. Changing this form the
            default (None) could have adverse effects!
        skip_rows : int | list | None
            Optional row index values to skip.

        Returns
        -------
        stdout : str
            HPC queue output string. Can be split on line breaks to get list.
        """
        job_name_str = ''
        if job_name is not None:
            job_name_str = ' -n {}'.format(job_name)

        if user is None:
            user = cls.USER

        if qformat is None:
            qformat = cls.SQ_FORMAT

        if skip_rows is None:
            skip_rows = cls.QSKIP

        cmd = ('squeue -u {user}{job_name} --format="{format_str}"'
               .format(user=user, job_name=job_name_str, format_str=qformat))
        stdout, _ = cls.submit(cmd)
        stdout = cls._skip_q_rows(stdout, skip_rows)

        return stdout

    @staticmethod
    def scontrol(cmd):
        """Submit an scontrol command.

        Parameters
        ----------
        cmd : str
            Command string after "scontrol" word
        """
        cmd = 'scontrol {}'.format(cmd)
        cmd = shlex.split(cmd)
        subprocess.call(cmd)

    def scancel(self, arg):
        """Cancel a slurm job.

        Parameters
        ----------
        arg : int | list | str
            SLURM integer job id(s) to cancel. Can be a list of integer
            job ids, 'all' to cancel all jobs, or a feature (-p short) to
            cancel all jobs with a given feature
        """

        if isinstance(arg, (list, tuple)):
            for job_id in arg:
                self.scancel(job_id)

        elif str(arg).lower() == 'all':
            self._queue = None
            for job_id in self.queue_job_ids:
                self.scancel(job_id)

        elif isinstance(arg, (int, str)):
            cmd = ('scancel {}'.format(arg))
            cmd = shlex.split(cmd)
            subprocess.call(cmd)

        else:
            e = ('Could not cancel: {} with type {}'
                 .format(arg, type(arg)))
            logger.error(e)
            raise ExecutionError(e)

    def change_qos(self, arg, qos):
        """Change the priority (quality of service) for a job.

        Parameters
        ----------
        arg : int | list | str
            SLURM integer job id(s) to change qos for.
            Can be 'all' for all jobs.
        qos : str
            New qos value
        """

        if isinstance(arg, (list, tuple)):
            for job_id in arg:
                self.change_qos(job_id, qos)

        elif isinstance(arg, int):
            cmd = 'update job {} QOS={}'.format(arg, qos)
            self.scontrol(cmd)

        elif str(arg).lower() == 'all':
            self._queue = None
            for job_id, attrs in self.queue.items():
                status = attrs[self.QCOL_STATUS].lower()
                if status == 'pd':
                    self.change_qos(job_id, qos)

        else:
            e = ('Could not change qos of: {} with type {}'
                 .format(arg, type(arg)))
            logger.error(e)
            raise ExecutionError(e)

    def hold(self, arg):
        """Temporarily hold a job from submitting. Held jobs will stay in queue
        but will not get nodes until released.

        Parameters
        ----------
        arg : int | list | str
            SLURM integer job id(s) to hold. Can be 'all' to hold all jobs.
        """

        if isinstance(arg, (list, tuple)):
            for job_id in arg:
                self.hold(job_id)

        elif isinstance(arg, int):
            cmd = 'hold {}'.format(arg)
            self.scontrol(cmd)

        elif str(arg).lower() == 'all':
            self._queue = None
            for job_id, attrs in self.queue.items():
                status = attrs[self.QCOL_STATUS].lower()
                if status == 'pd':
                    self.hold(job_id)

        else:
            e = ('Could not hold: {} with type {}'
                 .format(arg, type(arg)))
            logger.error(e)
            raise ExecutionError(e)

    def release(self, arg):
        """Release a job that was previously on hold so it will be submitted
        to a compute node.

        Parameters
        ----------
        arg : int | list | str
            SLURM integer job id(s) to release.
            Can be 'all' to release all jobs.
        """

        if isinstance(arg, (list, tuple)):
            for job_id in arg:
                self.release(job_id)

        elif isinstance(arg, int):
            cmd = 'release {}'.format(arg)
            self.scontrol(cmd)

        elif str(arg).lower() == 'all':
            self._queue = None
            for job_id, attrs in self.queue.items():
                status = attrs[self.QCOL_STATUS].lower()
                reason = attrs['NODELIST(REASON)'].lower()
                if status == 'pd' and 'jobheld' in reason:
                    self.release(job_id)

        else:
            e = ('Could not release: {} with type {}'
                 .format(arg, type(arg)))
            logger.error(e)
            raise ExecutionError(e)

    @staticmethod
    def _special_cmd_strs(feature, memory, module, module_root, conda_env):
        """Get special sbatch request strings for SLURM features, memory,
        modules, and conda environments

        Parameters
        ----------
        feature : str
            Additional flags for SLURM job. Format is "--qos=high"
            or "--depend=[state:job_id]". Default is None.
        memory : int
            Node memory request in GB.
        module : bool
            Module to load
        module_root : str
            Path to module root to load
        conda_env : str
            Conda environment to activate

        Returns
        -------
        feature_str : str
            SBATCH shell script feature request string.
        mem_str : str
            SBATCH shell script memory request string.
        env_str : str
            SBATCH shell script module load or source activate environment
            request string.
        """
        feature_str = ''
        if feature is not None:
            feature_str = '#SBATCH {}  # extra feature\n'.format(feature)

        mem_str = ''
        if memory is not None:
            mem_str = ('#SBATCH --mem={}  # node RAM in MB\n'
                       .format(int(memory * 1000)))

        env_str = ''
        if module is not None:
            env_str = ("echo module use {module_root}\n"
                       "module use {module_root}\n"
                       "echo module load {module}\n"
                       "module load {module}\n"
                       "echo module load complete!\n"
                       .format(module_root=module_root, module=module))
        elif conda_env is not None:
            env_str = ("echo source activate {conda_env}\n"
                       "source activate {conda_env}\n"
                       "echo conda env activate complete!\n"
                       .format(conda_env=conda_env))

        return feature_str, mem_str, env_str

    def sbatch(self, cmd, alloc, walltime, memory=None, feature=None,
               name='reV', stdout_path='./stdout', keep_sh=False,
               conda_env=None, module=None,
               module_root='/shared-projects/rev/modulefiles'):
        """Submit a SLURM job via sbatch command and SLURM shell script

        Parameters
        ----------
        cmd : str
            Command to be submitted in SLURM shell script. Example:
                'python -m reV.generation.cli_gen'
        alloc : str
            HPC project (allocation) handle. Example: 'rev'.
        walltime : float
            Node walltime request in hours.
        memory : int
            Node memory request in GB.
        feature : str
            Additional flags for SLURM job. Format is "--qos=high"
            or "--depend=[state:job_id]". Default is None.
        name : str
            SLURM job name.
        stdout_path : str
            Path to print .stdout and .stderr files.
        keep_sh : bool
            Boolean to keep the .sh files. Default is to remove these files
            after job submission.
        conda_env : str
            Conda environment to activate
        module : bool
            Module to load
        module_root : str
            Path to module root to load

        Returns
        -------
        out : str
            sbatch standard output, if submitted successfully, this is the
            slurm job id.
        err : str
            sbatch standard error, this is typically an empty string if the job
            was submitted successfully.
        """

        if len(name) > self.MAX_NAME_LEN:
            msg = ('Cannot submit job with name longer than {} chars: "{}"'
                   .format(self.MAX_NAME_LEN, name))
            logger.error(msg)
            raise ValueError(msg)

        status = self.check_status(job_name=name)

        if status is not None:
            logger.info('Not submitting job "{}" because it is in '
                        'squeue or has been recently submitted'.format(name))
            out = None
            err = 'already_running'

        else:
            self.make_path(stdout_path)
            special = self._special_cmd_strs(feature, memory, module,
                                             module_root, conda_env)
            fname = '{}.sh'.format(name)
            script = ('#!/bin/bash\n'
                      '#SBATCH --account={a}  # allocation account\n'
                      '#SBATCH --time={t}  # walltime\n'
                      '#SBATCH --job-name={n}  # job name\n'
                      '#SBATCH --nodes=1  # number of nodes\n'
                      '#SBATCH --output={p}/{n}_%j.o\n'
                      '#SBATCH --error={p}/{n}_%j.e\n{m}{f}'
                      'echo Running on: $HOSTNAME, Machine Type: $MACHTYPE\n'
                      '{e}\n{cmd}'
                      .format(a=alloc, t=self.format_walltime(walltime),
                              n=name, p=stdout_path, m=special[1],
                              f=special[0], e=special[2], cmd=cmd))

            # write the shell script file and submit as qsub job
            self.make_sh(fname, script)
            out, err = self.submit('sbatch {script}'.format(script=fname))

            if not keep_sh:
                self.rm(fname)

            if err:
                msg = 'Received a SLURM error or warning: {}'.format(err)
                logger.warning(msg)
                warn(msg, SlurmWarning)
            else:
                job_id = int(out.split(' ')[-1])
                out = str(job_id)
                logger.debug('SLURM job "{}" with id #{} submitted '
                             'successfully'.format(name, job_id))
                self._queue[job_id] = {self.QCOL_ID: job_id,
                                       self.QCOL_NAME: name,
                                       self.QCOL_STATUS: 'PD'}

        return out, err



# -*- coding: utf-8 -*-
"""
Custom dtypes for Click.
"""
import click
import logging


logger = logging.getLogger(__name__)


def sanitize_str(value, subs=('slice', '=', '(', ')', ' ')):
    """Sanitize characters from string."""
    for s in subs:
        value = value.replace(s, '')
    return value


class IntType(click.ParamType):
    """Integer click input argument type with option for None."""
    name = 'int'

    def convert(self, value, param, ctx):
        """Convert to int or return as None."""
        if isinstance(value, str):
            if 'None' in value:
                return None
            else:
                return int(value)
        elif isinstance(value, int):
            return value
        else:
            self.fail('Cannot recognize int type: {} {}'
                      .format(value, type(value)), param, ctx)


class FloatType(click.ParamType):
    """Float click input argument type with option for None."""
    name = 'float'

    def convert(self, value, param, ctx):
        """Convert to float or return as None."""
        if isinstance(value, str):
            if 'None' in value:
                return None
            else:
                return float(value)
        elif isinstance(value, float):
            return value
        else:
            self.fail('Cannot recognize float type: {} {}'
                      .format(value, type(value)), param, ctx)


class StrType(click.ParamType):
    """String click input argument type with option for None."""
    name = 'str'

    def convert(self, value, param, ctx):
        """Convert to str or return as None."""
        if isinstance(value, str):
            if value.lower() == 'none':
                return None
            else:
                return value
        else:
            self.fail('Cannot recognize str type: {} {}'
                      .format(value, type(value)), param, ctx)


class StrFloatType(click.ParamType):
    """String or float click input argument type with option for None."""
    name = 'strfloat'

    def convert(self, value, param, ctx):
        """Convert to str or float or return as None."""
        if isinstance(value, str):
            if value.lower() == 'none':
                return None
            else:
                try:
                    value = float(value)
                except ValueError:
                    pass
                return value
        else:
            self.fail('Cannot recognize str or float type: {} {}'
                      .format(value, type(value)), param, ctx)


class ListType(click.ParamType):
    """Base list click input argument type with option for None."""
    name = 'list'

    @staticmethod
    def dtype(x):
        """Option to enforce a Homogeneous datatype."""
        return x

    def convert(self, value, param, ctx):
        """Convert string to list."""
        if isinstance(value, str):
            value = sanitize_str(value, subs=['=', '(', ')', ' ', '[', ']',
                                              '"', "'"])
            if value == 'None':
                return None
            list0 = value.split(',')
            return [self.dtype(x) for x in list0]
        elif isinstance(value, list):
            return value
        elif isinstance(value, type(None)):
            return value
        else:
            self.fail('Cannot recognize list type: {} {}'
                      .format(value, type(value)), param, ctx)


class FloatListType(ListType):
    """Homogeneous list of floats click input argument type."""
    name = 'floatlist'

    @staticmethod
    def dtype(x):
        """Enforce a homogeneous float datatype."""
        return float(x)


class IntListType(ListType):
    """Homogeneous list of integers click input argument type."""
    name = 'intlist'

    @staticmethod
    def dtype(x):
        """Enforce a homogeneous integer datatype."""
        return int(x)


class StrListType(ListType):
    """Homogeneous list of strings click input argument type."""
    name = 'strlist'

    @staticmethod
    def dtype(x):
        """Enforce a homogeneous string datatype."""
        return str(x)


class PathListType(ListType):
    """Homogeneous list of click path input argument types."""
    name = 'pathlist'

    def __init__(self, **kwargs):
        """
        Parameters
        ----------
        kwargs : dict
            kwargs for click.types.Path
        """
        self._click_path_type = click.types.Path(**kwargs)

    def convert(self, value, param, ctx):
        """Convert string to list."""
        if isinstance(value, str):
            value = sanitize_str(value, subs=['=', '(', ')', ' ', '[', ']',
                                              '"', "'"])
            if value == 'None':
                return None
            list0 = value.split(',')
            return [self.path(x, param, ctx) for x in list0]
        elif isinstance(value, list):
            return value
        elif isinstance(value, type(None)):
            return value
        else:
            self.fail('Cannot recognize list type: {} {}'
                      .format(value, type(value)), param, ctx)

    def path(self, value, param, ctx):
        """Enforce a homogeneous string datatype."""
        return self._click_path_type.convert(value, param, ctx)


INT = IntType()
FLOAT = FloatType()
STR = StrType()
STRFLOAT = StrFloatType()
INTLIST = IntListType()
FLOATLIST = FloatListType()
STRLIST = StrListType()
PATHLIST = PathListType()



# -*- coding: utf-8 -*-
"""
Temporal Statistics Extraction
"""
from concurrent.futures import as_completed
import logging
import numpy as np
import os
import pandas as pd
from warnings import warn

from rex.rechunk_h5.rechunk_h5 import get_chunk_slices
from rex.resource import Resource
from rex.utilities.execution import SpawnProcessPool

logger = logging.getLogger(__name__)


class TemporalStats:
    """
    Temporal Statistics from Resource Data
    """
    STATS = {'mean': {'func': np.mean, 'kwargs': {'axis': 0}},
             'median': {'func': np.median, 'kwargs': {'axis': 0}},
             'std': {'func': np.std, 'kwargs': {'axis': 0}}}

    def __init__(self, res_h5, statistics='mean', res_cls=Resource,
                 hsds=False):
        """
        Parameters
        ----------
        res_h5 : str
            Path to resource h5 file(s)
        statistics : str | tuple | dict, optional
            Statistics to extract, either a key or tuple of keys in
            cls.STATS, or a dictionary of the form
            {'stat_name': {'func': *, 'kwargs: {**}}},
            by default 'mean'
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        """
        self._res_h5 = res_h5
        self._stats = None
        self.statistics = statistics

        self._res_cls = res_cls
        self._hsds = hsds

        with res_cls(res_h5, hsds=self._hsds) as f:
            self._time_index = f.time_index
            self._meta = f.meta

    @property
    def res_h5(self):
        """
        Path to resource h5 file(s)

        Returns
        -------
        str
        """
        return self._res_h5

    @property
    def statistics(self):
        """
        Dictionary of statistic functions/kwargs to run

        Returns
        -------
        dict
        """
        return self._stats

    @statistics.setter
    def statistics(self, statistics):
        """
         Statistics to extract, either a key or tuple of keys in
        cls.STATS, or a dictionary of the form
        {'stat_name': {'func': *, 'kwargs: {**}}}

        Parameters
        ----------
        statistics : dict
        """
        self._stats = self._check_stats(statistics)

    @property
    def res_cls(self):
        """
        Resource class to use to access res_h5

        Returns
        -------
        Class
        """
        return self._res_cls

    @property
    def time_index(self):
        """
        Resource Datetimes

        Returns
        -------
        pandas.DatetimeIndex
        """
        return self._time_index

    @property
    def meta(self):
        """
        Resource meta-data table

        Returns
        -------
        pandas.DataFrame
        """
        return self._meta

    @property
    def lat_lon(self):
        """
        Resource (lat, lon) coordinates

        Returns
        -------
        pandas.DataFrame
        """
        lat_lon_cols = ['latitude', 'longitude']
        for c in self.meta.columns:
            if c.lower() in ['lat', 'latitude']:
                lat_lon_cols[0] = c
            elif c.lower() in ['lon', 'long', 'longitude']:
                lat_lon_cols[1] = c

        return self.meta[lat_lon_cols]

    @classmethod
    def _check_stats(cls, statistics):
        """
        check desired statistics to make sure inputs are valid

        Parameters
        ----------
        statistics : str | tuple | dict
            Statistics to extract, either a key or tuple of keys in
            cls.STATS, or a dictionary of the form
            {'stat_name': {'func': *, 'kwargs: {**}}}

        Returns
        -------
        stats : dict
            Dictionary of statistic functions/kwargs to run
        """
        if isinstance(statistics, str):
            statistics = (statistics, )

        if isinstance(statistics, (tuple, list)):
            statistics = {s: cls.STATS[s] for s in statistics}

        for stat in statistics.values():
            msg = 'A "func"(tion) must be provided for each statistic'
            assert 'func' in stat, msg
            if 'kwargs' in stat:
                msg = 'statistic function kwargs must be a dictionary '
                assert isinstance(stat['kwargs'], dict), msg

        return statistics

    @staticmethod
    def _format_index_value(index, stat, month_map=None):
        """
        Format groupby index value

        Parameters
        ----------
        index : int | tuple
            hour, month, or (month, hour) groupby index value
        stat : str
            Statistic that was computed
        month_map : dict | None, optional
            Mapping of month int to str, by default None

        Returns
        -------
        out : str

        """
        if isinstance(index, np.ndarray):
            m, h = index
            if month_map is not None:
                m = month_map[m]
            else:
                m = '{:02d}'.format(m)

            out = "{}-{:02d}".format(m, h)
        else:
            if month_map is not None:
                out = month_map[index]
            else:
                out = '{:02d}'.format(index)

        out += '_{}'.format(stat)

        return out

    @classmethod
    def _create_names(cls, index, stat):
        """
        Generate statistics names

        Parameters
        ----------
        index : pandas.Index | pandas.MultiIndex
            Temporal index, either month, hour, or (month, hour)
        stat : str
            Statistic that was computed

        Returns
        -------
        columns : list
            List of column names to use
        """
        month_map = {1: 'Jan', 2: 'Feb', 3: 'Mar', 4: 'Apr', 5: 'May',
                     6: 'June', 7: 'July', 8: 'Aug', 9: 'Sept', 10: 'Oct',
                     11: 'Nov', 12: 'Dec'}
        index = np.array(index.to_list())
        if len(index.shape) != 2 and index.max() > 12:
            month_map = None

        columns = [cls._format_index_value(i, stat, month_map=month_map)
                   for i in index]

        return columns

    @classmethod
    def _compute_stats(cls, res_data, statistics, diurnal=False, month=False):
        """
        Compute desired stats for desired time intervals from res_data

        Parameters
        ----------
        res_data : pandas.DataFrame
            DataFrame or resource data. Index is time_index, columns are sites
        statistics : dict
            Dictionary of statistic functions/kwargs to run
        diurnal : bool, optional
            Extract diurnal stats, by default False
        month : bool, optional
            Extract monthly stats, by default False

        Returns
        -------
        res_stats : pandas.DataFrame
            DataFrame of desired statistics at desired time intervals
        """
        groupby = []
        if month:
            groupby.append(res_data.index.month)

        if diurnal:
            groupby.append(res_data.index.hour)

        if groupby:
            res_data = res_data.groupby(groupby)

        res_stats = []
        for name, stat in statistics.items():
            func = stat['func']
            kwargs = stat.get('kwargs', {})
            s_data = res_data.aggregate(func, **kwargs)

            if groupby:
                columns = cls._create_names(s_data.index, name)
                s_data = s_data.T
                s_data.columns = columns
            else:
                s_data = s_data.to_frame(name=name)

            res_stats.append(s_data)

        res_stats = pd.concat(res_stats, axis=1)

        return res_stats

    @classmethod
    def _extract_stats(cls, res_h5, res_cls, statistics, dataset, hsds=False,
                       time_index=None, sites_slice=None, diurnal=False,
                       month=False, combinations=False):
        """
        Extract stats for given dataset, sites, and temporal extent

        Parameters
        ----------
        res_h5 : str
            Path to resource h5 file(s)
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        statistics : dict
            Dictionary of statistic functions/kwargs to run
        dataset : str
            Dataset to extract stats for
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        time_index : pandas.DatatimeIndex | None, optional
            Resource DatetimeIndex, if None extract from res_h5,
            by default None
        sites_slice : slice | None, optional
            Sites to extract, if None all, by default None
        diurnal : bool, optional
            Extract diurnal stats, by default False
        month : bool, optional
            Extract monthly stats, by default False
        combinations : bool, optional
            Extract all combinations of temporal stats, by default False

        Returns
        -------
        res_stats : pandas.DataFrame
            DataFrame of desired statistics at desired time intervals
        """
        if sites_slice is None:
            sites_slice = slice(None, None, None)

        with res_cls(res_h5, hsds=hsds) as f:
            if time_index is None:
                time_index = f.time_index

            res_data = pd.DataFrame(f[dataset, :, sites_slice],
                                    index=time_index)
        if combinations:
            res_stats = [cls._compute_stats(res_data, statistics)]
            if month:
                res_stats.append(cls._compute_stats(res_data, statistics,
                                                    month=True))

            if diurnal:
                res_stats.append(cls._compute_stats(res_data, statistics,
                                                    diurnal=True))
            if month and diurnal:
                res_stats.append(cls._compute_stats(res_data, statistics,
                                                    month=True, diurnal=True))

            res_stats = pd.concat(res_stats, axis=1)
        else:
            res_stats = cls._compute_stats(res_data, statistics,
                                           diurnal=diurnal, month=month)

        if isinstance(sites_slice, slice) and sites_slice.stop:
            res_stats.index = \
                list(range(*sites_slice.indices(sites_slice.stop)))
        elif isinstance(sites_slice, (list, np.ndarray)):
            res_stats.index = sites_slice

        res_stats.index.name = 'gid'

        return res_stats

    @staticmethod
    def _slice_sites(sites_slice, n_sites, slice_size):
        """
        Break up sites_slice into slices of size slice_size

        Parameters
        ----------
        sites_slice : slice
            Sites to extract as a slice object to extract
        n_sites : int
            Total number of sites to extract
        slice_size : int
            Number of sites in each slice to extract either on each worker,
            or in series

        Returns
        -------
        slices : list
            List of slices to extract
        """
        stop = sites_slice.stop
        if stop is None:
            stop = n_sites

        step = sites_slice.step
        if step is not None:
            slice_size *= step

        if slice_size >= n_sites:
            msg = ('The slice_size {} is >= the number of sites to be '
                   'extracted {}! A single slice will be extracted.'
                   .format(slice_size, n_sites))
            logger.warning(msg)
            warn(msg)
            slices = [sites_slice]
        else:
            # Create slices of size slice_size
            slices = [slice(s, e, step) for s, e
                      in get_chunk_slices(stop, slice_size)]

        return slices

    @staticmethod
    def _split_sites(sites, slice_size):
        """
        Split sites into sub-lists of ~ size slice_size

        Parameters
        ----------
        sites : slice
            Sites to extract as a list or numpy object to extract
        slice_size : int
            Number of sites in each slice to extract either on each worker,
            or in series

        Returns
        -------
        slices : list
            List of slices to extract
        """
        if slice_size >= len(sites):
            msg = ('The slice_size {} is >= the number of sites to be '
                   'extracted {}! A single slice will be extracted.'
                   .format(slice_size, len(sites)))
            logger.warning(msg)
            warn(msg)
            slices = [sites]
        else:
            slices = np.array_split(sites, len(sites) // slice_size)

        return slices

    def _get_slices(self, dataset, sites=None, chunks_per_slice=5):
        """
        Get slices to extract

        Parameters
        ----------
        dataset : str
            Dataset to extract data from
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
        chunks_per_slice : int, optional
            Number of chunks to extract in each slice, by default 5

        Returns
        -------
        slices : list
            List of slices to extract
        """
        with self.res_cls(self.res_h5) as f:
            shape, _, chunks = f.get_dset_properties(dataset)

        if len(shape) != 2:
            msg = ('Cannot extract temporal stats for dataset {}, as it is '
                   'not a timeseries dataset!'.format(dataset))
            logger.error(msg)
            raise RuntimeError(msg)

        if chunks is not None:
            slice_size = chunks[1] * chunks_per_slice
        else:
            slice_size = chunks_per_slice

        if sites is None:
            sites = slice(None)

        if isinstance(sites, slice):
            slices = self._slice_sites(sites, shape[1], slice_size)
        elif isinstance(sites, (list, tuple, np.ndarray)):
            slices = self._split_sites(sites, slice_size)
        else:
            msg = ('sites must be of type "None", "slice", "list", "tuple", '
                   'or "np.ndarray", but {} was provided'.format(type(sites)))
            raise TypeError(msg)

        return slices

    def compute_statistics(self, dataset, sites=None,
                           diurnal=False, month=False, combinations=False,
                           max_workers=None, chunks_per_worker=5,
                           lat_lon_only=True):
        """
        Compute statistics

        Parameters
        ----------
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
        diurnal : bool, optional
            Extract diurnal stats, by default False
        month : bool, optional
            Extract monthly stats, by default False
        combinations : bool, optional
            Extract all combinations of temporal stats, by default False
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_slice : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True

        Returns
        -------
        res_stats : pandas.DataFrame
            DataFrame of desired statistics at desired time intervals
        """
        if max_workers is None:
            max_workers = os.cpu_count()

        slices = self._get_slices(dataset, sites,
                                  chunks_per_slice=chunks_per_worker)
        if len(slices) == 1:
            max_workers = 1

        if max_workers > 1:
            msg = ('Extracting {} for {} in parallel using {} workers'
                   .format(self.statistics.keys(), dataset, max_workers))
            logger.info(msg)

            loggers = [__name__, 'rex']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = []
                for sites_slice in slices:
                    future = exe.submit(self._extract_stats,
                                        self.res_h5, self.res_cls,
                                        self.statistics, dataset,
                                        hsds=self._hsds,
                                        time_index=self.time_index,
                                        sites_slice=sites_slice,
                                        diurnal=diurnal,
                                        month=month,
                                        combinations=combinations)
                    futures.append(future)

                res_stats = []
                for i, future in enumerate(as_completed(futures)):
                    res_stats.append(future.result())
                    logger.debug('Completed {} out of {} workers'
                                 .format((i + 1), len(futures)))

            res_stats = pd.concat(res_stats)
        else:
            msg = ('Extracting {} for {} in serial'
                   .format(self.statistics.keys(), dataset))
            logger.info(msg)
            if chunks_per_worker is not None:
                res_stats = []
                for sites_slice in slices:
                    res_stats.append(self._extract_stats(
                        self.res_h5, self.res_cls, self.statistics, dataset,
                        hsds=self._hsds, time_index=self.time_index,
                        sites_slice=sites_slice, diurnal=diurnal, month=month,
                        combinations=combinations))

                res_stats = pd.concat(res_stats)
            else:
                res_stats = self._extract_stats(
                    self.res_h5, self.res_cls, self.statistics, dataset,
                    hsds=self._hsds, time_index=self.time_index,
                    diurnal=diurnal, month=month,
                    combinations=combinations)

        if lat_lon_only:
            meta = self.lat_lon
        else:
            meta = self.meta

        res_stats = meta.join(res_stats.sort_index(), how='inner')

        return res_stats

    def full_stats(self, dataset, sites=None, max_workers=None,
                   chunks_per_worker=5, lat_lon_only=True):
        """
        Compute stats for entire temporal extent of file

        Parameters
        ----------
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_slice : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True

        Returns
        -------
        full_stats : pandas.DataFrame
            DataFrame of statistics for the entire temporal extent of file
        """
        full_stats = self.compute_statistics(
            dataset, sites=sites,
            max_workers=max_workers,
            chunks_per_worker=chunks_per_worker,
            lat_lon_only=lat_lon_only)

        return full_stats

    def monthly_stats(self, dataset, sites=None, max_workers=None,
                      chunks_per_worker=5, lat_lon_only=True):
        """
        Compute monthly stats

        Parameters
        ----------
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_slice : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True

        Returns
        -------
        monthly_stats : pandas.DataFrame
            DataFrame of monthly statistics
        """
        monthly_stats = self.compute_statistics(
            dataset, sites=sites, month=True,
            max_workers=max_workers,
            chunks_per_worker=chunks_per_worker,
            lat_lon_only=lat_lon_only)

        return monthly_stats

    def diurnal_stats(self, dataset, sites=None, max_workers=None,
                      chunks_per_worker=5, lat_lon_only=True):
        """
        Compute diurnal stats

        Parameters
        ----------
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_slice : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True

        Returns
        -------
        diurnal_stats : pandas.DataFrame
            DataFrame of diurnal statistics
        """
        diurnal_stats = self.compute_statistics(
            dataset, sites=sites, diurnal=True,
            max_workers=max_workers,
            chunks_per_worker=chunks_per_worker,
            lat_lon_only=lat_lon_only)

        return diurnal_stats

    def monthly_diurnal_stats(self, dataset, sites=None,
                              max_workers=None, chunks_per_worker=5,
                              lat_lon_only=True):
        """
        Compute monthly-diurnal stats

        Parameters
        ----------
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_slice : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True

        Returns
        -------
        monthly_diurnal_stats : pandas.DataFrame
            DataFrame of monthly-diurnal statistics
        """
        diurnal_stats = self.compute_statistics(
            dataset, sites=sites, month=True, diurnal=True,
            max_workers=max_workers,
            chunks_per_worker=chunks_per_worker,
            lat_lon_only=lat_lon_only)

        return diurnal_stats

    def all_stats(self, dataset, sites=None, max_workers=None,
                  chunks_per_worker=5, lat_lon_only=True):
        """
        Compute annual, monthly, monthly-diurnal, and diurnal stats

        Parameters
        ----------
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_slice : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True

        Returns
        -------
        all_diurnal_stats : pandas.DataFrame
            DataFrame of temporal statistics
        """
        all_stats = self.compute_statistics(
            dataset, sites=sites, month=True, diurnal=True, combinations=True,
            max_workers=max_workers,
            chunks_per_worker=chunks_per_worker,
            lat_lon_only=lat_lon_only)

        return all_stats

    def save_stats(self, res_stats, out_path):
        """
        Save statistics to disk

        Parameters
        ----------
        res_stats : pandas.DataFrame
            Table of statistics to save
        out_path : str
            Directory, .csv, or .json path to save statistics too
        """
        if os.path.isdir(out_path):
            out_fpath = os.path.splitext(os.path.basename(self.res_h5))[0]
            out_fpath = os.path.join(out_path, out_fpath + '.csv')
        else:
            out_fpath = out_path

        if out_fpath.endswith('.csv'):
            res_stats.to_csv(out_fpath)
        elif out_fpath.endswith('.json'):
            res_stats.to_json(out_fpath)
        else:
            msg = ("Cannot save statistics, expecting a directory, .csv, or "
                   ".json path, but got: {}".format(out_path))
            logger.error(msg)
            raise OSError(msg)

    @classmethod
    def run(cls, res_h5, dataset, sites=None, statistics='mean',
            diurnal=False, month=False, combinations=False,
            res_cls=Resource, hsds=False, max_workers=None,
            chunks_per_worker=5, lat_lon_only=True, out_path=None):
        """
        Compute temporal stats, by default full temporal extent stats

        Parameters
        ----------
        res_h5 : str
            Path to resource h5 file(s)
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
        statistics : str | tuple, optional
            Statistics to extract, must be 'mean', 'median', 'std',
            and/or 'stdev', by default 'mean'
        diurnal : bool, optional
            Extract diurnal stats, by default False
        month : bool, optional
            Extract monthly stats, by default False
        combinations : bool, optional
            Extract all combinations of temporal stats, by default False
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_slice : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        out_path : str, optional
            Directory, .csv, or .json path to save statistics too,
            by default None

        Returns
        -------
        annual_stats : pandas.DataFrame
            DataFrame of annual statistics
        """
        res_stats = cls(res_h5, statistics=statistics, res_cls=res_cls,
                        hsds=hsds)
        annual_stats = res_stats.compute_statistics(
            dataset, sites=sites,
            diurnal=diurnal, month=month, combinations=combinations,
            max_workers=max_workers, chunks_per_worker=chunks_per_worker,
            lat_lon_only=lat_lon_only)
        if out_path is not None:
            res_stats.save_stats(annual_stats, out_path)

        return annual_stats

    @classmethod
    def monthly(cls, res_h5, dataset, sites=None, statistics='mean',
                res_cls=Resource, hsds=False, max_workers=None,
                chunks_per_worker=5, lat_lon_only=True, out_path=None):
        """
        Compute monthly stats

        Parameters
        ----------
        res_h5 : str
            Path to resource h5 file(s)
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
        statistics : str | tuple, optional
            Statistics to extract, must be 'mean', 'median', 'std',
            and/or 'stdev', by default 'mean'
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_slice : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        out_path : str, optional
            Directory, .csv, or .json path to save statistics too,
            by default None

        Returns
        -------
        monthly_stats : pandas.DataFrame
            DataFrame of monthly statistics
        """
        monthly_stats = cls.run(res_h5, dataset, sites=sites,
                                statistics=statistics, diurnal=False,
                                month=True, combinations=False,
                                res_cls=res_cls, hsds=hsds,
                                max_workers=max_workers,
                                chunks_per_worker=chunks_per_worker,
                                lat_lon_only=lat_lon_only, out_path=out_path)

        return monthly_stats

    @classmethod
    def diurnal(cls, res_h5, dataset, sites=None, statistics='mean',
                res_cls=Resource, hsds=False, max_workers=None,
                chunks_per_worker=5, lat_lon_only=True, out_path=None):
        """
        Compute diurnal stats

        Parameters
        ----------
        res_h5 : str
            Path to resource h5 file(s)
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
        statistics : str | tuple, optional
            Statistics to extract, must be 'mean', 'median', 'std',
            and/or 'stdev', by default 'mean'
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_slice : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        out_path : str, optional
            Directory, .csv, or .json path to save statistics too,
            by default None

        Returns
        -------
        diurnal_stats : pandas.DataFrame
            DataFrame of diurnal statistics
        """
        diurnal_stats = cls.run(res_h5, dataset, sites=sites,
                                statistics=statistics, diurnal=True,
                                month=False, combinations=False,
                                res_cls=res_cls, hsds=hsds,
                                max_workers=max_workers,
                                chunks_per_worker=chunks_per_worker,
                                lat_lon_only=lat_lon_only, out_path=out_path)

        return diurnal_stats

    @classmethod
    def monthly_diurnal(cls, res_h5, dataset, sites=None,
                        statistics='mean', res_cls=Resource, hsds=False,
                        max_workers=None, chunks_per_worker=5,
                        lat_lon_only=True, out_path=None):
        """
        Compute monthly-diurnal stats

        Parameters
        ----------
        res_h5 : str
            Path to resource h5 file(s)
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
        statistics : str | tuple, optional
            Statistics to extract, must be 'mean', 'median', 'std',
            and/or 'stdev', by default 'mean'
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_slice : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        out_path : str, optional
            Directory, .csv, or .json path to save statistics too,
            by default None

        Returns
        -------
        monthly_diurnal_stats : pandas.DataFrame
            DataFrame of monthly-diurnal statistics
        """
        monthly_diurnal_stats = cls.run(res_h5, dataset, sites=sites,
                                        statistics=statistics, diurnal=True,
                                        month=True, combinations=False,
                                        res_cls=res_cls, hsds=hsds,
                                        max_workers=max_workers,
                                        chunks_per_worker=chunks_per_worker,
                                        lat_lon_only=lat_lon_only,
                                        out_path=out_path)

        return monthly_diurnal_stats

    @classmethod
    def all(cls, res_h5, dataset, sites=None, statistics='mean',
            res_cls=Resource, hsds=False, max_workers=None,
            chunks_per_worker=5, lat_lon_only=True, out_path=None):
        """
        Compute annual, monthly, monthly-diurnal, and diurnal stats

        Parameters
        ----------
        res_h5 : str
            Path to resource h5 file(s)
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
        statistics : str | tuple, optional
            Statistics to extract, must be 'mean', 'median', 'std',
            and/or 'stdev', by default 'mean'
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_slice : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        out_path : str, optional
            Directory, .csv, or .json path to save statistics too,
            by default None

        Returns
        -------
        all_stats : pandas.DataFrame
            DataFrame of temporal statistics
        """
        all_stats = cls.run(res_h5, dataset, sites=sites,
                            statistics=statistics, diurnal=True,
                            month=True, combinations=True,
                            res_cls=res_cls, hsds=hsds,
                            max_workers=max_workers,
                            chunks_per_worker=chunks_per_worker,
                            lat_lon_only=lat_lon_only, out_path=out_path)

        return all_stats



# -*- coding: utf-8 -*-
"""
Resource Extraction Tools
"""
import logging
import numpy as np
import os
import pandas as pd
import pickle
from scipy.spatial import cKDTree
from tempfile import TemporaryDirectory

from rex.multi_file_resource import (MultiFileNSRDB, MultiFileResource,
                                     MultiFileWTK)
from rex.multi_time_resource import MultiTimeResource
from rex.multi_year_resource import MultiYearResource
from rex.resource import Resource
from rex.renewable_resource import (NSRDB, SolarResource, WaveResource,
                                    WindResource)
from rex.resource_extraction.temporal_stats import TemporalStats
from rex.utilities.exceptions import ResourceValueError
from rex.utilities.utilities import parse_year, check_tz

TREE_DIR = TemporaryDirectory()
logger = logging.getLogger(__name__)


class ResourceX:
    """
    Resource data extraction tool
    """
    def __init__(self, res_h5, res_cls=Resource, tree=None, unscale=True,
                 hsds=False, str_decode=True, group=None):
        """
        Parameters
        ----------
        res_h5 : str
            Path to resource .h5 file of interest
        res_cls : obj
            Resource class to use to open and access resource data
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        group : str
            Group within .h5 resource file to open
        """
        self._res = res_cls(res_h5, unscale=unscale, hsds=hsds,
                            str_decode=str_decode, group=group)
        self._tree = tree
        self._i = 0

    def __repr__(self):
        msg = "{} extractor for {}".format(self._res.__class__.__name__,
                                           self.resource.h5_file)

        return msg

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

        if type is not None:
            raise

    def __len__(self):
        return self.h5['meta'].shape[0]

    def __getitem__(self, keys):
        return self.resource[keys]

    def __iter__(self):
        return self

    def __next__(self):
        if self._i >= len(self.datasets):
            self._i = 0
            raise StopIteration

        dset = self.datasets[self._i]
        self._i += 1

        return dset

    def __contains__(self, dset):
        return dset in self.datasets

    @property
    def resource(self):
        """
        Open res_cls instance to access res_h5 data

        Returns
        -------
        res_cls
        """
        return self._res

    @property
    def h5(self):
        """
        Open h5py File instance. If _group is not None return open Group

        Returns
        -------
        h5 : h5py.File | h5py.Group
            Open h5py File or Group instance
        """
        return self.resource.h5

    @property
    def datasets(self):
        """
        Datasets available

        Returns
        -------
        list
            List of datasets
        """
        return self.resource.datasets

    @property
    def dsets(self):
        """
        Datasets available

        Returns
        -------
        list
            List of datasets
        """
        return self.datasets

    @property
    def groups(self):
        """
        Groups available

        Returns
        -------
        groups : list
            List of groups
        """
        return self.resource.groups

    @property
    def shape(self):
        """
        Resource shape (timesteps, sites)
        shape = (len(time_index), len(meta))

        Returns
        -------
        shape : tuple
            Shape of resource variable arrays (timesteps, sites)
        """
        return self.resource.shape

    @property
    def meta(self):
        """
        Meta data DataFrame

        Returns
        -------
        meta : pandas.DataFrame
            Resource Meta Data
        """
        return self.resource.meta

    @property
    def time_index(self):
        """
        DatetimeIndex

        Returns
        -------
        time_index : pandas.DatetimeIndex
            Resource datetime index
        """
        return self.resource.time_index

    @property
    def coordinates(self):
        """
        Coordinates: (lat, lon) pairs

        Returns
        -------
        lat_lon : ndarray
            Array of (lat, lon) pairs for each site in meta
        """
        return self.resource.lat_lon

    @property
    def lat_lon(self):
        """
        Extract (latitude, longitude) pairs

        Returns
        -------
        lat_lon : ndarray
        """
        return self.resource.lat_lon

    @property
    def global_attrs(self):
        """
        Global (file) attributes

        Returns
        -------
        global_attrs : dict
        """
        return self.resource.attrs

    @property
    def attrs(self):
        """
        Global (file) attributes

        Returns
        -------
        attrs : dict
            .h5 file attributes sourced from first .h5 file
        """
        return self.global_attrs

    @property
    def tree(self):
        """
        Returns
        -------
        tree : cKDTree
            Lat, lon coordinates cKDTree
        """
        if not isinstance(self._tree, cKDTree):
            self._tree = self._init_tree(self._tree)

        return self._tree

    @property
    def countries(self):
        """
        Returns
        -------
        countries : ndarray
            Countries available in .h5 file
        """
        if 'country' in self.meta:
            countries = self.meta['country'].unique()
        else:
            countries = None

        return countries

    @property
    def states(self):
        """
        Returns
        -------
        states : ndarray
            States available in .h5 file
        """
        if 'state' in self.meta:
            states = self.meta['state'].unique()
        else:
            states = None

        return states

    @property
    def counties(self):
        """
        Returns
        -------
        counties : ndarray
            Counties available in .h5 file
        """
        if 'county' in self.meta:
            counties = self.meta['county'].unique()
        else:
            counties = None

        return counties

    @staticmethod
    def _get_tree_file(h5_file):
        """
        Create path to pre-computed tree from h5_file by splitting file name
        at year if available, else replacing the .h5 suffix

        Parameters
        ----------
        h5_file : str
            Path to source .h5 file

        Returns
        -------
        tree_file : str
            Path to pre-comupted tree .pkl file name
        """
        f_name = os.path.basename(h5_file)
        try:
            year = parse_year(f_name)
            tree_file = f_name.split(str(year))[0] + 'tree.pkl'
        except RuntimeError:
            tree_file = f_name.replace('.h5', '_tree.pkl')

        return tree_file

    @staticmethod
    def _load_tree(tree_path):
        """
        Load tree from pickle file

        Parameters
        ----------
        tree_path : str
            Pickle (.pkl, .pickle) file containing precomputed cKDTree

        Returns
        -------
        tree : cKDTree
            Precomputed tree of lat, lon coordinates
        """
        try:
            with open(tree_path, 'rb') as f:
                tree = pickle.load(f)
        except Exception as e:
            logger.warning('Could not extract tree from {}: {}'
                           .format(tree_path, e))
            tree = None

        return tree

    @staticmethod
    def _save_tree(tree, tree_path):
        """
        Save pre-computed Tree to TEMP_DIR as a pickle file

        Parameters
        ----------
        tree : cKDTree
            pre-computed cKDTree
        tree_path : str
            Path to pickle file in TEMP_DIR to save tree too
        """
        try:
            with open(tree_path, 'wb') as f:
                pickle.dump(tree, f)
        except Exception as e:
            logger.warning('Could not save tree to {}: {}'
                           .format(tree_path, e))

    @staticmethod
    def _to_SAM_csv(sam_df, site_meta, out_path):
        """
        Save SAM dataframe to disk and add meta data to header to make
        SAM compliant

        Parameters
        ----------
        sam_df : pandas.DataFrame
            rex SAM DataFrame
        site_meta : pandas.DataFrame
            Site meta data
        out_path : str
            Path to .csv file to save data too
        """
        if not out_path.endswith('.csv'):
            if os.path.isfile(out_path):
                out_path = os.path.basename(out_path)

            out_path = os.path.join(out_path, "{}.csv".format(sam_df.name))

        sam_df.to_csv(out_path, index=False)

        if 'gid' not in site_meta:
            site_meta.index.name = 'gid'
            site_meta = site_meta.reset_index()

        col_map = {}
        for c in site_meta.columns:
            if c == 'timezone':
                col_map[c] = 'Time Zone'
            elif c == 'gid':
                col_map[c] = 'Location ID'
            else:
                col_map[c] = c.capitalize()

        site_meta = site_meta.rename(columns=col_map)
        cols = ','.join(site_meta.columns)
        values = ','.join(site_meta.values[0].astype(str))

        with open(out_path, 'r+') as f:
            content = f.read()
            f.seek(0, 0)
            f.write(cols + '\n' + values + '\n' + content)

    def _init_tree(self, tree):
        """
        Inititialize cKDTree of lat, lon coordinates

        Parameters
        ----------
        tree : str | cKDTree | NoneType
            Path to .pgz file containing pre-computed tree
            If None search bin for .pgz file matching h5 file
            else compute tree

        Returns
        -------
        tree : cKDTree
            cKDTree of lat, lon coordinate from wtk .h5 file
        """
        tree_path = self._get_tree_file(self.resource.h5_file)
        if not isinstance(tree, (cKDTree, str, type(None))):
            tree = None
            logger.warning('Precomputed tree must be supplied as a pickle '
                           'file or a cKDTree, not a {}'
                           .format(type(tree)))

        if tree is None:
            if tree_path in os.listdir(TREE_DIR.name):
                tree = os.path.join(TREE_DIR.name, tree_path)

        if isinstance(tree, str):
            tree = self._load_tree(tree)

        if tree is None:
            lat_lon = self.lat_lon
            tree = cKDTree(lat_lon)  # pylint: disable=not-callable
            self._save_tree(tree, os.path.join(TREE_DIR.name, tree_path))

        return tree

    def _check_lat_lon(self, lat_lon):
        """
        Check lat lon coordinates against domain

        Parameters
        ----------
        lat_lon : ndarray
            Either a single (lat, lon) pair or series of (lat, lon) pairs
        """
        lat_min, lat_max = self.lat_lon[:, 0].min(), self.lat_lon[:, 0].max()
        lon_min, lon_max = self.lat_lon[:, 1].min(), self.lat_lon[:, 1].max()

        if not isinstance(lat_lon, np.ndarray):
            lat_lon = np.array(lat_lon)

        if len(lat_lon.shape) == 1:
            lat_lon = np.expand_dims(lat_lon, axis=0)

        check = lat_lon[:, 0] < lat_min
        check |= lat_lon[:, 0] > lat_max
        check |= lat_lon[:, 1] < lon_min
        check |= lat_lon[:, 1] > lon_max

        if any(check):
            bad_coords = lat_lon[check]
            msg = ("Latitude, longitude coordinates ({}) are outsides of the "
                   "resource domain: (({}, {}), ({}, {}))"
                   .format(bad_coords, lat_min, lon_min, lat_max, lon_max))
            raise ResourceValueError(msg)

        return lat_lon

    def lat_lon_gid(self, lat_lon):
        """
        Get nearest gid to given (lat, lon) pair or pairs

        Parameters
        ----------
        lat_lon : ndarray
            Either a single (lat, lon) pair or series of (lat, lon) pairs

        Returns
        -------
        gids : int | ndarray
            Nearest gid(s) to given (lat, lon) pair(s)
        """
        lat_lon = self._check_lat_lon(lat_lon)
        _, gids = self.tree.query(lat_lon)

        if len(gids) == 1:
            gids = int(gids[0])

        return gids

    def region_gids(self, region, region_col='state'):
        """
        Get the gids for given region

        Parameters
        ----------
        region : str
            Region to search for
        region_col : str
            Region column to search

        Returns
        -------
        gids : ndarray
            Vector of gids in given region
        """
        gids = self.meta
        gids = gids[gids[region_col] == region].index.values

        return gids

    def box_gids(self, lat_lon_1, lat_lon_2):
        """
        Get gids within bounding lat_lon coordinates

        Parameters
        ----------
        lat_lon_1 : list | tuple
            One corner of the bounding box
        lat_lon_2 : list | tuple
            The other corner of the bounding box

        Returns
        -------
        gids : ndarray
            Gids in bounding box
        """
        self._check_lat_lon(np.vstack((lat_lon_1, lat_lon_2)))
        lat_min, lat_max = sorted([lat_lon_1[0], lat_lon_2[0]])
        lon_min, lon_max = sorted([lat_lon_1[1], lat_lon_2[1]])

        coords = self.lat_lon
        gids = coords[:, 0] >= lat_min
        gids &= coords[:, 0] <= lat_max
        gids &= coords[:, 1] >= lon_min
        gids &= coords[:, 1] <= lon_max

        return np.where(gids)[0]

    def timestep_idx(self, timestep):
        """
        Get the index of the desired timestep

        Parameters
        ----------
        timestep : str
            Timestep of interest

        Returns
        -------
        ts_idx : int
            Time index value
        """
        timestep = check_tz(pd.to_datetime(timestep))
        idx = np.where(self.time_index == timestep)[0][0]

        return idx

    def get_gid_ts(self, ds_name, gid):
        """
        Extract timeseries of site(s) neareset to given lat_lon(s)

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        gid : int | list
            Resource gid(s) of interset

        Return
        ------
        ts : ndarray
            Time-series for given site(s) and dataset
        """
        ts = self[ds_name, :, gid]

        return ts

    def get_gid_df(self, ds_name, gid):
        """
        Extract timeseries of site(s) nearest to given lat_lon(s) and return
        as a DataFrame

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        gid : int | list
            Resource gid(s) of interset

        Return
        ------
        df : pandas.DataFrame
            Time-series DataFrame for given site(s) and dataset
        """
        index = pd.Index(data=self.time_index, name='time_index')
        if isinstance(gid, int):
            df = pd.DataFrame(self[ds_name, :, gid], columns=[gid],
                              index=index)
            df.name = gid
        else:
            df = pd.DataFrame(self[ds_name, :, gid], columns=gid,
                              index=index)
            df.name = ds_name

        return df

    def get_lat_lon_ts(self, ds_name, lat_lon):
        """
        Extract timeseries of site(s) neareset to given lat_lon(s)

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        lat_lon : tuple | list
            (lat, lon) coordinate of interest or pairs of coordinates

        Return
        ------
        ts : ndarray
            Time-series for given site(s) and dataset
        """
        gid = self.lat_lon_gid(lat_lon)
        ts = self.get_gid_ts(ds_name, gid)

        return ts

    def get_lat_lon_df(self, ds_name, lat_lon):
        """
        Extract timeseries of site(s) nearest to given lat_lon(s) and return
        as a DataFrame

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        lat_lon : tuple
            (lat, lon) coordinate of interest

        Return
        ------
        df : pandas.DataFrame
            Time-series DataFrame for given site(s) and dataset
        """
        gid = self.lat_lon_gid(lat_lon)
        df = self.get_gid_df(ds_name, gid)

        return df

    def get_region_ts(self, ds_name, region, region_col='state'):
        """
        Extract timeseries of of all sites in given region

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        region : str
            Region to search for
        region_col : str
            Region column to search

        Return
        ------
        region_ts : ndarray
            Time-series array of desired dataset for all sites in desired
            region
        """
        gids = self.region_gids(region, region_col=region_col)
        region_ts = self.get_gid_ts(ds_name, gids)

        return region_ts

    def get_region_df(self, ds_name, region, region_col='state'):
        """
        Extract timeseries of of all sites in given region and return as a
        DataFrame

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        region : str
            Region to extract all pixels for
        region_col : str
            Region column to search

        Return
        ------
        region_df : pandas.DataFrame
            Time-series array of desired dataset for all sites in desired
            region
        """
        gids = self.region_gids(region, region_col=region_col)
        region_df = self.get_gid_df(ds_name, gids)

        return region_df

    def get_box_ts(self, ds_name, lat_lon_1, lat_lon_2):
        """
        Extract timeseries of of all sites in given bounding box

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        lat_lon_1 : list | tuple
            One corner of the bounding box
        lat_lon_2 : list | tuple
            The other corner of the bounding box

        Return
        ------
        box_ts : ndarray
            Time-series array of desired dataset for all sites in desired
            bounding box
        """
        gids = self.box_gids(lat_lon_1, lat_lon_2)
        box_ts = self.get_gid_ts(ds_name, gids)

        return box_ts

    def get_box_df(self, ds_name, lat_lon_1, lat_lon_2):
        """
        Extract timeseries of of all sites in given bounding box and return as
        a DataFrame

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        lat_lon_1 : list | tuple
            One corner of the bounding box
        lat_lon_2 : list | tuple
            The other corner of the bounding box

        Return
        ------
        box_df : pandas.DataFrame
            Time-series array of desired dataset for all sites in desired
            bounding box
        """
        gids = self.box_gids(lat_lon_1, lat_lon_2)
        box_df = self.get_gid_df(ds_name, gids)

        return box_df

    def get_SAM_gid(self, gid, out_path=None, **kwargs):
        """
        Extract time-series of all variables needed to run SAM for nearest
        site to given resource gid

        Parameters
        ----------
        gid : int | list
            Resource gid(s) of interset
        out_path : str, optional
            Path to save SAM data to in SAM .csv format, by default None
        kwargs : dict
            Internal kwargs for _get_SAM_df

        Return
        ------
        SAM_df : pandas.DataFrame | list
            Time-series DataFrame for given site and dataset
            If multiple lat, lon pairs are given a list of DatFrames is
            returned
        """
        if isinstance(gid, int):
            gid = [gid, ]

        SAM_df = []
        for res_id in gid:
            # pylint: disable=E1111
            df = self.resource._get_SAM_df('SAM', res_id, **kwargs)
            SAM_df.append(df)
            if out_path is not None:
                site_meta = self['meta', res_id]
                self._to_SAM_csv(df, site_meta, out_path)

        if len(SAM_df) == 1:
            SAM_df = SAM_df[0]

        return SAM_df

    def get_SAM_lat_lon(self, lat_lon, out_path=None, **kwargs):
        """
        Extract time-series of all variables needed to run SAM for nearest
        site to given lat_lon

        Parameters
        ----------
        lat_lon : tuple
            (lat, lon) coordinate of interest
        out_path : str, optional
            Path to save SAM data to in SAM .csv format, by default None
        kwargs : dict
            Internal kwargs for _get_SAM_df

        Return
        ------
        SAM_df : pandas.DataFrame | list
            Time-series DataFrame for given site and dataset
            If multiple lat, lon pairs are given a list of DatFrames is
            returned
        """
        gid = self.lat_lon_gid(lat_lon)
        SAM_df = self.get_SAM_gid(gid, out_path=out_path, **kwargs)

        return SAM_df

    def get_timestep_map(self, ds_name, timestep, region=None,
                         region_col='state', box=None):
        """
        Extract a map of the given dataset at the given timestep for the
        given region if supplied

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        timestep : str
            Timestep of interest
        region : str, optional
            Region to extract all pixels for, by default None
        region_col : str, optional
            Region column to search, by default 'state'
        box : tuple, optional
            Bounding corners of box to extract pixels for

        Returns
        -------
        ts_map : pandas.DataFrame
            DataFrame of map values
        """
        lat_lons = self.lat_lon
        ts_idx = self.timestep_idx(timestep)
        gids = slice(None)

        if region is not None and box is not None:
            msg = 'Can only process a region OR a set of box corners!'
            raise RuntimeError(msg)

        if region is not None:
            gids = self.region_gids(region, region_col=region_col)
            lat_lons = lat_lons[gids]
        elif box is not None:
            gids = self.box_gids(*box)
            lat_lons = lat_lons[gids]

        ts_map = self[ds_name, ts_idx, gids]
        ts_map = pd.DataFrame({'longitude': lat_lons[:, 1],
                               'latitude': lat_lons[:, 0],
                               ds_name: ts_map})

        return ts_map

    def close(self):
        """
        Close res_cls instance
        """
        self._res.close()


class MultiFileResourceX(ResourceX):
    """
    Multi-File resource extraction class
    """

    def __init__(self, resource_path, res_cls=MultiFileResource, tree=None,
                 unscale=True, str_decode=True, check_files=False):
        """
        Parameters
        ----------
        resource_path : str
            Path to resource .h5 files
            Available formats:
                /h5_dir/
                /h5_dir/prefix*suffix
        res_cls : obj
            Resource class to use to open and access resource data
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        check_files : bool
            Check to ensure files have the same coordinates and time_index
        """
        self._res = res_cls(resource_path, unscale=unscale,
                            str_decode=str_decode, check_files=check_files)
        self._tree = tree
        self._i = 0


class MultiYearResourceX(ResourceX):
    """
    Multi Year resource extraction class
    """
    def __init__(self, resource_path, years=None, tree=None, unscale=True,
                 str_decode=True, hsds=False, res_cls=Resource):
        """
        Parameters
        ----------
        resource_path : str
            Path to resource .h5 files
            Available formats:
                /h5_dir/
                /h5_dir/prefix*suffix
        years : list, optional
            List of years to access, by default None
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        res_cls : obj
            Resource handler to use to open individual .h5 files
        """
        self._res = MultiYearResource(resource_path, years=years,
                                      unscale=unscale, str_decode=str_decode,
                                      hsds=hsds, res_cls=res_cls)
        self._tree = tree
        self._i = 0

    def get_means_map(self, ds_name, year=None, region=None,
                      region_col='state', max_workers=None,
                      chunks_per_worker=5):
        """
        Extract given year(s) and compute means

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        year : str | list, optional
            Year(s) to compute means for, by default None
        region : str
            Region to extract all pixels for
        region_col : str
            Region column to search
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_slice : int, optional
            Number of chunks to extract on each worker, by default 5

        Returns
        -------
        ts_map : pandas.DataFrame
            DataFrame of map values
        """
        gids = slice(None)
        if region is not None:
            gids = self.region_gids(region, region_col=region_col)

        if year is None:
            year = slice(None)

        means_map = TemporalStats.run(self.resource.h5_file, ds_name,
                                      sites=gids, statistics='mean',
                                      res_cls=self.resource.__class__,
                                      hsds=self.resource.hsds,
                                      max_workers=max_workers,
                                      chunks_per_worker=chunks_per_worker,
                                      lat_lon_only=True)

        return means_map


class MultiTimeResourceX(ResourceX):
    """
    Resource extraction class for data stored temporaly accross multiple files
    """

    def __init__(self, resource_path, tree=None, unscale=True,
                 str_decode=True, hsds=False, res_cls=Resource):
        """
        Parameters
        ----------
        resource_path : str
            Path to resource .h5 files
            Available formats:
                /h5_dir/
                /h5_dir/prefix*suffix
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        res_cls : obj
            Resource handler to us to open individual .h5 files
        """
        self._res = MultiTimeResource(resource_path, unscale=unscale,
                                      str_decode=str_decode, hsds=hsds,
                                      res_cls=res_cls)
        self._tree = tree
        self._i = 0


class SolarX(ResourceX):
    """
    Solar Resource extraction class
    """
    def __init__(self, solar_h5, tree=None, unscale=True, hsds=False,
                 str_decode=True, group=None):
        """
        Parameters
        ----------
        solar_h5 : str
            Path to solar .h5 file of interest
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        group : str
            Group within .h5 resource file to open
        """
        super().__init__(solar_h5, unscale=unscale, str_decode=str_decode,
                         group=group, hsds=hsds, tree=tree,
                         res_cls=SolarResource)


class NSRDBX(ResourceX):
    """
    NSRDB extraction class
    """
    def __init__(self, nsrdb_h5, tree=None, unscale=True, hsds=False,
                 str_decode=True, group=None):
        """
        Parameters
        ----------
        nsrdb_h5 : str
            Path to NSRDB .h5 file of interest
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        group : str
            Group within .h5 resource file to open
        """
        super().__init__(nsrdb_h5, unscale=unscale, str_decode=str_decode,
                         group=group, hsds=hsds, tree=tree,
                         res_cls=NSRDB)


class MultiFileNSRDBX(MultiFileResourceX):
    """
    Multi-File NSRDB extraction class
    """
    def __init__(self, nsrdb_source, tree=None, unscale=True, str_decode=True,
                 check_files=False):
        """
        Parameters
        ----------
        nsrdb_source : str | list
            Path to directory containing multi-file NSRDB file sets.
            Available formats:
                /h5_dir/
                /h5_dir/prefix*suffix
            Or list of source NSRDB .h5 files
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        check_files : bool
            Check to ensure files have the same coordinates and time_index
        """
        super().__init__(nsrdb_source, unscale=unscale, str_decode=str_decode,
                         check_files=check_files, tree=tree,
                         res_cls=MultiFileNSRDB)


class MultiYearNSRDBX(MultiYearResourceX):
    """
    Multi Year NSRDB extraction class
    """
    def __init__(self, nsrdb_path, years=None, tree=None, unscale=True,
                 str_decode=True, hsds=False):
        """
        Parameters
        ----------
        nsrdb_path : str
            Path to NSRDB .h5 files
            Available formats:
                /h5_dir/
                /h5_dir/prefix*suffix
        years : list, optional
            List of years to access, by default None
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        """
        super().__init__(nsrdb_path, years=years, tree=tree, unscale=unscale,
                         str_decode=str_decode, hsds=hsds, res_cls=NSRDB)


class MultiTimeNSRDBX(MultiTimeResourceX):
    """
    NSRDB extraction class for data stored temporaly accross multiple files
    """

    def __init__(self, nsrdb_path, tree=None, unscale=True,
                 str_decode=True, hsds=False):
        """
        Parameters
        ----------
        nsrdb_path : str
            Path to NSRDB .h5 files
            Available formats:
                /h5_dir/
                /h5_dir/prefix*suffix
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        """
        super().__init__(nsrdb_path, tree=tree, unscale=unscale,
                         str_decode=str_decode, hsds=hsds, res_cls=NSRDB)


class WindX(ResourceX):
    """
    Wind Resource extraction class
    """
    def __init__(self, wind_h5, tree=None, unscale=True, hsds=False,
                 str_decode=True, group=None):
        """
        Parameters
        ----------
        wind_h5 : str
            Path to Wind .h5 file of interest
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        group : str
            Group within .h5 resource file to open
        """
        super().__init__(wind_h5, unscale=unscale, str_decode=str_decode,
                         group=group, hsds=hsds, tree=tree,
                         res_cls=WindResource)

    def get_SAM_gid(self, hub_height, gid, out_path=None, **kwargs):
        """
        Extract time-series of all variables needed to run SAM for nearest
        site to given resource gid and hub height

        Parameters
        ----------
        hub_height : int
            Hub height of interest
        gid : int | list
            Resource gid(s) of interset
        out_path : str, optional
            Path to save SAM data to in SAM .csv format, by default None
        kwargs : dict
            Internal kwargs for _get_SAM_df:
            - require_wind_dir
            - icing

        Return
        ------
        SAM_df : pandas.DataFrame | list
            Time-series DataFrame for given site and dataset
            If multiple lat, lon pairs are given a list of DatFrames is
            returned
        """
        ds_name = 'SAM_{}m'.format(hub_height)
        if isinstance(gid, int):
            gid = [gid, ]

        SAM_df = []
        for res_id in gid:
            df = self.resource._get_SAM_df(ds_name, res_id, **kwargs)
            SAM_df.append(df)
            if out_path is not None:
                site_meta = self['meta', res_id]
                self._to_SAM_csv(df, site_meta, out_path)

        if len(SAM_df) == 1:
            SAM_df = SAM_df[0]

        return SAM_df

    def get_SAM_lat_lon(self, hub_height, lat_lon, out_path=None, **kwargs):
        """
        Extract time-series of all variables needed to run SAM for nearest
        site to given lat_lon and hub height

        Parameters
        ----------
        hub_height : int
            Hub height of interest
        gid : int | list
            Resource gid(s) of interset
        out_path : str, optional
            Path to save SAM data to in SAM .csv format, by default None
        kwargs : dict
            Internal kwargs for _get_SAM_df:
            - require_wind_dir
            - icing

        Return
        ------
        SAM_df : pandas.DataFrame | list
            Time-series DataFrame for given site and dataset
            If multiple lat, lon pairs are given a list of DatFrames is
            returned
        """
        gid = self.lat_lon_gid(lat_lon)
        SAM_df = self.get_SAM_gid(hub_height, gid, out_path=out_path, **kwargs)

        return SAM_df


class MultiFileWindX(MultiFileResourceX):
    """
    Multi-File Wind Resource extraction class
    """
    def __init__(self, wtk_source, tree=None, unscale=True, str_decode=True,
                 check_files=False):
        """
        Parameters
        ----------
        wtk_source : str | list
            Path to directory containing multi-file WTK file sets.
            Available formats:
                /h5_dir/
                /h5_dir/prefix*suffix
            Or list of source WTK .h5 files
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        check_files : bool
            Check to ensure files have the same coordinates and time_index
        """
        super().__init__(wtk_source, unscale=unscale, str_decode=str_decode,
                         check_files=check_files, tree=tree,
                         res_cls=MultiFileWTK)


class MultiYearWindX(MultiYearResourceX):
    """
    Multi Year Wind Resource extraction class
    """
    def __init__(self, wtk_path, years=None, tree=None, unscale=True,
                 str_decode=True, hsds=False):
        """
        Parameters
        ----------
        wtk_path : str
            Path to annual WTK .h5 files
            Available formats:
                /h5_dir/
                /h5_dir/prefix*suffix
        years : list, optional
            List of years to access, by default None
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        """
        super().__init__(wtk_path, years=years, tree=tree, unscale=unscale,
                         str_decode=str_decode, hsds=hsds,
                         res_cls=WindResource)


class MultiTimeWindX(MultiTimeResourceX):
    """
    Wind resource extraction class for data stored temporaly accross multiple
    files
    """

    def __init__(self, wtk_path, tree=None, unscale=True, str_decode=True,
                 hsds=False):
        """
        Parameters
        ----------
        wtk_path : str
            Path to annual WTK .h5 files
            Available formats:
                /h5_dir/
                /h5_dir/prefix*suffix
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        """
        super().__init__(wtk_path, tree=tree, unscale=unscale,
                         str_decode=str_decode, hsds=hsds,
                         res_cls=WindResource)


class WaveX(ResourceX):
    """
    Wave data extraction class
    """

    def __init__(self, wave_h5, tree=None, unscale=True, hsds=False,
                 str_decode=True, group=None):
        """
        Parameters
        ----------
        wave_h5 : str
            Path to US_Wave .h5 file of interest
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        group : str
            Group within .h5 resource file to open
        """
        super().__init__(wave_h5, unscale=unscale, str_decode=str_decode,
                         group=group, hsds=hsds, tree=tree,
                         res_cls=WaveResource)

    def get_gid_ts(self, ds_name, gid):
        """
        Extract timeseries of site(s) neareset to given lat_lon(s)

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        gid : int | list
            Resource gid(s) of interset

        Return
        ------
        ts : ndarray
            Time-series for given site(s) and dataset
        """
        if ds_name == 'directional_wave_spectrum':
            ts = self[ds_name, :, :, :, gid]
        else:
            ts = self[ds_name, :, gid]

        return ts

    def get_gid_df(self, ds_name, gid):
        """
        Extract timeseries of site(s) nearest to given lat_lon(s) and return
        as a DataFrame

        Parameters
        ----------
        ds_name : str
            Dataset to extract
        gid : int | list
            Resource gid(s) of interset

        Return
        ------
        df : pandas.DataFrame
            Time-series DataFrame for given site(s) and dataset
        """
        if ds_name == 'directional_wave_spectrum':
            df = self[ds_name, :, :, :, gid]
            index = pd.MultiIndex.from_product(
                [self.time_index, self['frequency'], self['direction']],
                names=['time_index', 'frequency', 'direction'])
            ax1 = np.product(df.shape[:3])
            ax2 = df.shape[-1] if len(df.shape) == 4 else 1
            df = df.reshape(ax1, ax2)
        else:
            df = self[ds_name, :, gid]
            index = pd.Index(data=self.time_index, name='time_index')

        if isinstance(gid, int):
            df = pd.DataFrame(df, columns=[gid],
                              index=index)
            df.name = gid
        else:
            df = pd.DataFrame(df, columns=gid,
                              index=index)
            df.name = ds_name

        return df


class MultiYearWaveX(MultiYearResourceX):
    """
    Multi Year Wave extraction class
    """

    def __init__(self, wave_path, years=None, tree=None, unscale=True,
                 str_decode=True, hsds=False):
        """
        Parameters
        ----------
        wave_path : str
            Path to US_Wave .h5 files
            Available formats:
                /h5_dir/
                /h5_dir/prefix*suffix
        years : list, optional
            List of years to access, by default None
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        """
        super().__init__(wave_path, years=years, tree=tree, unscale=unscale,
                         str_decode=str_decode, hsds=hsds,
                         res_cls=WaveResource)


class MultiTimeWaveX(MultiTimeResourceX):
    """
    Wave resource extraction class for data stored temporaly accross multiple
    files
    """

    def __init__(self, wave_path, tree=None, unscale=True, str_decode=True,
                 hsds=False):
        """
        Parameters
        ----------
        wave_path : str
            Path to US_Wave .h5 files
            Available formats:
                /h5_dir/
                /h5_dir/prefix*suffix
        tree : str | cKDTree
            cKDTree or path to .pkl file containing pre-computed tree
            of lat, lon coordinates
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        hsds : bool
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS
        """
        super().__init__(wave_path, tree=tree, unscale=unscale,
                         str_decode=str_decode, hsds=hsds,
                         res_cls=WindResource)



# -*- coding: utf-8 -*-
"""
Logging wrapper
"""
import logging
import sys
import os
import psutil
import time

FORMAT = '%(levelname)s - %(asctime)s [%(filename)s:%(lineno)d] : %(message)s'
LOG_LEVEL = {'INFO': logging.INFO,
             'DEBUG': logging.DEBUG,
             'WARNING': logging.WARNING,
             'ERROR': logging.ERROR,
             'CRITICAL': logging.CRITICAL}


def create_dirs(dir_path):
    """
    Wrapper for checking and making directory to handle parallel processing

    Parameters
    ----------
    dir_path : str
        Directory path to create if it doesn't exist
    """
    i = 0
    while True:
        try:
            if not os.path.exists(dir_path):
                os.makedirs(dir_path)

            break
        except FileExistsError:
            time.sleep(0.01)
        except Exception as ex:
            i += 1
            if i == 2:
                raise FileExistsError('Cannot create {}: {}'
                                      .format(dir_path, ex))


def get_handler(log_level="INFO", log_file=None, log_format=FORMAT):
    """
    get logger handler

    Parameters
    ----------
    log_level : str
        handler-specific logging level, must be key in LOG_LEVEL.
    log_file : str
        path to the log file
    log_format : str
        format string to use with the logging package

    Returns
    -------
    handler : logging.FileHandler | logging.StreamHandler
        handler to add to logger
    """
    if log_file:
        # file handler with mode "a"
        handler = logging.FileHandler(log_file, mode='a')
    else:
        # stream handler to system stdout
        handler = logging.StreamHandler(sys.stdout)

    if log_format:
        logformat = logging.Formatter(log_format)
        handler.setFormatter(logformat)

    # Set a handler-specific logging level (root logger should be at debug)
    handler.setLevel(LOG_LEVEL[log_level.upper()])

    return handler


def setup_logger(logger_name, stream=True, log_level="INFO", log_file=None,
                 log_format=FORMAT):
    """
    Setup logging instance with given name and attributes

    Parameters
    ----------
    logger_name : str
        Name of logger
    stream : bool
        Init stream logger
    log_level : str
        Level of logging to capture, must be key in LOG_LEVEL. If multiple
        handlers/log_files are requested in a single call of this function,
        the specified logging level will be applied to all requested handlers.
    log_file : str | list
        Path to file to use for logging, if None use a StreamHandler
        list of multiple handlers is permitted
    log_format : str
        Format for loggings, default is FORMAT

    Returns
    -------
    logger : logging.logger
        instance of logger for given name, with given level and added handler
    handler : logging.FileHandler | logging.StreamHandler | list
        handler(s) added to logger
    """
    logger = logging.getLogger(logger_name)
    current_handlers = [str(h) for h in logger.handlers]

    # Set root logger to debug, handlers will control levels above debug
    logger.setLevel(LOG_LEVEL["DEBUG"])

    handlers = []
    if isinstance(log_file, list):
        for h in log_file:
            handlers.append(get_handler(log_level=log_level, log_file=h,
                                        log_format=log_format))
    else:
        handlers.append(get_handler(log_level=log_level, log_file=log_file,
                                    log_format=log_format))

    if stream:
        if log_file is not None:
            handlers.append(get_handler())

    for handler in handlers:
        if str(handler) not in current_handlers:
            logger.addHandler(handler)

    return logger


class LoggingAttributes:
    """
    Class to store and pass logging attributes to modules
    """
    def __init__(self):
        self._loggers = {}

    def __setitem__(self, logger_name, attributes):
        log_attrs = self[logger_name]
        for attr, value in attributes.items():
            if attr == 'log_file':
                handlers = list(log_attrs.get('log_file', []))
                if not isinstance(value, (list, tuple)):
                    # make the log_file request into a iterable list
                    value = [value]

                for v in value:
                    if v not in handlers:
                        # check if each handler has been previously set
                        handlers.append(v)

                log_attrs[attr] = handlers
            else:
                log_attrs[attr] = value

        self._loggers[logger_name] = log_attrs

    def __getitem__(self, logger_name):
        return self._loggers.get(logger_name, {})

    def init_logger(self, loggers):
        """
        Extract logger attributes and initialize logger

        Parameters
        ----------
        loggers : str | list
            Logger names to initialize
        """
        if not isinstance(loggers, (list, tuple)):
            loggers = [loggers]

        for logger_name in loggers:
            try:
                attrs = self[logger_name]
                setup_logger(logger_name, **attrs)
            except KeyError:
                pass


LOGGERS = LoggingAttributes()


def init_logger(logger_name, stream=True, log_level="INFO", log_file=None,
                log_format=FORMAT):
    """
    Starts logging instance and adds logging attributes to REV_LOGGERS

    Parameters
    ----------
    logger_name : str
        Name of logger to initialize
    log_level : str
        Level of logging to capture, must be key in LOG_LEVEL. If multiple
        handlers/log_files are requested in a single call of this function,
        the specified logging level will be applied to all requested handlers.
    log_file : str | list
        Path to file to use for logging, if None use a StreamHandler
        list of multiple handlers is permitted
    log_format : str
        Format for loggings, default is FORMAT

    Returns
    -------
    logger : logging.logger
        logging instance that was initialized
    """
    kwargs = {"log_level": log_level, "log_file": log_file,
              "log_format": log_format, 'stream': stream}
    logger = setup_logger(logger_name, **kwargs)

    LOGGERS[logger_name] = kwargs

    return logger


def init_mult(name, logdir, modules, verbose=False, node=False):
    """Init multiple loggers to a single file or stdout.

    Parameters
    ----------
    name : str
        Job name; name of log file.
    logdir : str
        Target directory to save .log files.
    modules : list | tuple
        List of reV modules to initialize loggers for.
    verbose : bool
        Option to turn on debug logging.
    node : bool
        Flag for whether this is a node-level logger. If this is a node logger,
        and the log level is info, the log_file will be None (sent to stdout).

    Returns
    -------
    loggers : list
        List of logging instances that were initialized.
    """

    if verbose:
        log_level = 'DEBUG'
    else:
        log_level = 'INFO'

    if not os.path.exists(logdir):
        os.makedirs(logdir)

    loggers = []
    for module in modules:
        log_file = os.path.join(logdir, '{}.log'.format(name))

        # check for redundant loggers in the REV_LOGGERS singleton
        logger = LOGGERS[module]

        if ((not node or (node and log_level == 'DEBUG'))
                and 'log_file' not in logger):
            # No log file belongs to this logger, init a logger file
            logger = init_logger(module, log_level=log_level,
                                 log_file=log_file)
        elif node and log_level == 'INFO':
            # Node level info loggers only go to STDOUT/STDERR files
            logger = init_logger(module, log_level=log_level, log_file=None)

        loggers.append(logger)

    return loggers


def log_mem(logger, log_level='DEBUG'):
    """Log the memory usage to the input logger object.

    Parameters
    ----------
    logger : logging.Logger
        Logger object to log memory message to.
    log_level : str
        DEBUG or INFO for different log levels for this log message.

    Returns
    -------
    msg : str
        Memory utilization log message string.
    """
    mem = psutil.virtual_memory()
    msg = ('Memory utilization is {0:.3f} GB out of '
           '{1:.3f} GB total ({2:.1f}% used)'
           .format(mem.used / 1e9, mem.total / 1e9,
                   100 * mem.used / mem.total))
    if log_level.upper() == 'DEBUG':
        logger.debug('\t- {}'.format(msg))
    else:
        logger.info(msg)

    return msg
